/*****************************************************************************/
/*  file_transfer.c - contains the ui and network routines for file transfer */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

extern struct ftp_transfer_data *file_transfers;
extern pthread_mutex_t transfer_mutex;
extern struct ftp_window_data window1, window2;

static struct ftp_transfer_data *inittrans(struct ftp_window_data *fromwdw, struct ftp_window_data *towdw, int direction);
void dotrans(struct ftp_transfer_data *tdata);
static void asktrans(struct ftp_transfer_data *tdata);
static void restCB(GtkWidget *widget, struct ftp_transfer_data *tdata);
static void ovrCB(GtkWidget *widget, struct ftp_transfer_data *tdata);
static void restallCB(GtkWidget *widget, struct ftp_transfer_data *tdata);
static void ovrallCB(GtkWidget *widget, struct ftp_transfer_data *tdata);
static void cancel(GtkWidget *widget, struct ftp_transfer_data *tdata);
 
void retrCB(GtkWidget *widget, gpointer data) {
   struct ftp_transfer_data *tdata;
   
   if(window2.local == -1) {
      ftp_log(LOG_MISC, "Retrieve Files: Not connected to a remote site\n");
      return;
   }
   else if(window2.numselected == 0) {
      ftp_log(LOG_MISC, "Retrieve Files: You must have at least one item selected\n");
      return;
   }
   tdata = inittrans(&window2, &window1, 1);
   dotrans(tdata);
}
/*****************************************************************************/
void retrmenu(gpointer data) {
   retrCB(NULL, data);
}
/*****************************************************************************/
void putCB(GtkWidget *widget, gpointer data) {
   struct ftp_transfer_data *tdata;
   
   if(window2.local == -1) {
      ftp_log(LOG_MISC, "Put Files: Not connected to a remote site\n");
      return;
   }
   else if(window1.numselected == 0) {
      ftp_log(LOG_MISC, "Put Files: You must have at least one item selected\n");
      return;
   }
   tdata = inittrans(&window1, &window2, 0);
   dotrans(tdata);
}
/*****************************************************************************/
void putmenu(gpointer data) {
   putCB(NULL, data);
}
/*****************************************************************************/
static struct ftp_transfer_data *inittrans(struct ftp_window_data *fromwdw, struct ftp_window_data *towdw, int direction) {
   struct ftp_file_data *newfle, *tempfle;
   struct ftp_transfer_data *tdata, *temptdata;
   char dest[MAXSTR];
   int i;
   
   pthread_mutex_lock(&transfer_mutex);
   tdata = create_new_tdata();
   if(direction) tdata->flags |= TRANSFER_DIRECTION;
   tdata->wdata = fromwdw;
   tdata->otherwdata = towdw;
   if(tdata->flags & TRANSFER_DIRECTION) {
      *tdata->hdata = *fromwdw->host;
      strncpy(dest, towdw->host->dir, sizeof(dest));
   }
   else {
      *tdata->hdata = *towdw->host;
      strncpy(dest, fromwdw->host->dir, sizeof(dest));
   }
   dest[sizeof(dest)-1] = '\0';
   tdata->hdata->totalfiles = tdata->wdata->numselected;
   tdata->hdata->last = NULL;
   tempfle = tdata->wdata->host->files;
   for(i=0; i<tdata->wdata->totalitems; i++) {
      if(!tdata->wdata->items[i]) {
         tempfle = tempfle->next;
         continue;
      }
      newfle = mymalloc(sizeof(struct ftp_file_data));
      *newfle = *tempfle;
      g_snprintf(newfle->file, sizeof(newfle->file), "%s/%s", dest, tempfle->file);
      newfle->file[sizeof(newfle->file)-1] = '\0';
      newfle->flags = 0;
      newfle->next = NULL;
      if(tdata->hdata->last == NULL) tdata->hdata->files = newfle;
      else tdata->hdata->last->next = newfle;
      tdata->hdata->last = newfle;
      tempfle = tempfle->next;
   }
   tdata->curfle = tdata->hdata->files;
   tdata->next = NULL;
   if(file_transfers == NULL) file_transfers = tdata;
   else {
      temptdata = file_transfers;
      while(temptdata->next != NULL) temptdata = temptdata->next;
      temptdata->next = tdata;
   }
   pthread_mutex_unlock(&transfer_mutex);
   return(tdata);
}
/*****************************************************************************/
struct ftp_transfer_data *create_new_tdata(void) {
   struct ftp_transfer_data *tdata;

   tdata = mymalloc(sizeof(struct ftp_transfer_data));
   tdata->hdata = mymalloc(sizeof(struct ftp_host_data));
   *tdata->hdata->recvdata.recvdata = '\0';
   tdata->hdata->recvdata.pos = tdata->hdata->recvdata.recvdata;
   tdata->flags = 0;
   tdata->flags |= TRANSFER_NEW; /* Signal to make this listbox entry */
   tdata->curtrans = 0;
   tdata->starttime = time(NULL);
   return(tdata);
}
/*****************************************************************************/
void dotrans(struct ftp_transfer_data *tdata) {
   struct ftp_file_data *tempfle;
   struct stat flestat;
   char *pos;
   
   if(tdata->flags & TRANSFER_DOALL) {
      pthread_mutex_lock(&transfer_mutex);
      while(tdata->curfle != NULL) {
         if((tdata->flags & TRANSFER_DIRECTION) && stat(tdata->curfle->file, &flestat) != -1
            && (tdata->flags & TRANSFER_OP)) {

            tdata->curfle->flags |= FILE_RESTART;
            tdata->curfle->startsize = flestat.st_size;
         }
         else if(!(tdata->flags & TRANSFER_DIRECTION) && (tdata->flags & TRANSFER_OP)) {
            /* If we are uploading the file, see if it is in the other listbox */
            pos = strrchr(tdata->curfle->file, '/');
            if(pos == NULL) pos = tdata->curfle->file;
            else pos++;
            tempfle = tdata->otherwdata->host->files;
            while(tempfle != NULL) {
               if(strcmp(tempfle->file, pos) == 0) {
                  tdata->curfle->flags |= FILE_RESTART;
                  break;
               }
               tempfle = tempfle->next;
            }
         }
         tdata->curfle = tdata->curfle->next;
      }
      tdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
      tdata->flags |= TRANSFER_SHOW;
      pthread_mutex_unlock(&transfer_mutex);
   }
   else {
      pthread_mutex_lock(&transfer_mutex);
      while(tdata->curfle != NULL) {
         if((tdata->flags & TRANSFER_DIRECTION) && stat(tdata->curfle->file, &flestat) != -1) {
            tdata->curfle->startsize = flestat.st_size;
            break;
         }
         else if(!(tdata->flags & TRANSFER_DIRECTION)) {
            /* If we are uploading the file, see if it is in the other listbox */
            pos = strrchr(tdata->curfle->file, '/');
            if(pos == NULL) pos = tdata->curfle->file;
            else pos++;
            tempfle = tdata->otherwdata->host->files;
            while(tempfle != NULL) {
               if(strcmp(tempfle->file, pos) == 0) break; 
               tempfle = tempfle->next;
            }
            if(tempfle != NULL) break;
            else tdata->curfle = tdata->curfle->next;
         }
         else tdata->curfle = tdata->curfle->next;
      }         
      if(tdata->curfle == NULL) {
         tdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
         tdata->flags |= TRANSFER_SHOW;
      }
      pthread_mutex_unlock(&transfer_mutex);
      if(tdata->curfle != NULL) asktrans(tdata);
   }
}
/*****************************************************************************/
static void asktrans(struct ftp_transfer_data *tdata) {
   char tempstr[MAXSTR], *pos;

   pos = strrchr(tdata->curfle->file, '/');
   if(pos == NULL) pos = tdata->curfle->file;
   else pos++;
   g_snprintf(tempstr, sizeof(tempstr), "File %s already exists on %s\nWhat do you want to do?", 
      pos, tdata->flags & TRANSFER_DIRECTION ? "local computer" : "remote computer");
   MakeYesNoDialog("Transfer Files", tempstr, 5,
      "Overwrite", ovrCB, tdata,
      "Resume", restCB, tdata,
      "Overwrite All", ovrallCB, tdata,
      "Resume All", restallCB, tdata,
      "Cancel", cancel, tdata);
}
/*****************************************************************************/
static void restCB(GtkWidget *widget, struct ftp_transfer_data *tdata) {
   pthread_mutex_lock(&transfer_mutex);
   tdata->curfle->flags |= FILE_RESTART;;
   tdata->curfle = tdata->curfle->next;
   pthread_mutex_unlock(&transfer_mutex);
   dotrans(tdata);
}
/*****************************************************************************/
static void ovrCB(GtkWidget *widget, struct ftp_transfer_data *tdata) {
   pthread_mutex_lock(&transfer_mutex);
   tdata->curfle = tdata->curfle->next;
   pthread_mutex_unlock(&transfer_mutex);
   dotrans(tdata);
}
/*****************************************************************************/
static void restallCB(GtkWidget *widget, struct ftp_transfer_data *tdata) {
   pthread_mutex_lock(&transfer_mutex);
   tdata->flags |= TRANSFER_OP | TRANSFER_DOALL;
   pthread_mutex_unlock(&transfer_mutex);
   dotrans(tdata);
}
/*****************************************************************************/
static void ovrallCB(GtkWidget *widget, struct ftp_transfer_data *tdata) {
   pthread_mutex_lock(&transfer_mutex);
   tdata->flags |= TRANSFER_DOALL;
   pthread_mutex_unlock(&transfer_mutex);
   dotrans(tdata);
}
/*****************************************************************************/
static void cancel(GtkWidget *widget, struct ftp_transfer_data *tdata) {
   pthread_mutex_lock(&transfer_mutex);
   tdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
   tdata->flags |= TRANSFER_DONE;
   pthread_mutex_unlock(&transfer_mutex);
}
/*****************************************************************************/
void *ftp_get_files(void *ptr) {
   char tempstr[MAXSTR], buf[8192], gotstr[MAXSTR], ofstr[MAXSTR], *pos;
   struct ftp_transfer_data *tdata;
   int datafd, i, j=0;
   FILE *writefd;
   float kbs;
   ssize_t n;
   int num;

   num = 0;
   tdata = (struct ftp_transfer_data *) ptr;
   pthread_detach(pthread_self());
   if(ftp_connect(tdata->wdata, tdata->hdata, tdata)) {
      if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, NULL, 0, "TYPE %c\r\n", tdata->hdata->type == FTP_ASCII ? 'A' : 'I') != '2') {
         queue_log(LOG_MISC, "Error: Could not set %s mode\n", tdata->hdata->type == FTP_ASCII ? "ascii" : "binary");
      }
      else while(!(tdata->flags & TRANSFER_CANCEL) && tdata->curfle != NULL) {
         num++;
         if((datafd = ftp_init_data_conn(tdata->wdata, tdata->hdata)) == 0) break;

         if(tdata->flags & TRANSFER_DONE_DO_VIEW) {
            writefd = fopen(tdata->dest, tdata->curfle->flags & FILE_RESTART ? "a" : "w");
         }
         else {
            writefd = fopen(tdata->curfle->file, tdata->curfle->flags & FILE_RESTART ? "a" : "w");
         }
         if(!writefd) {
            queue_log(LOG_MISC, "Cannot create %s: %s\n", tdata->curfle->file , g_strerror(errno));
            break;
         }
         if(tdata->curfle->flags & FILE_RESTART) {
            if(tdata->hdata->type == FTP_ASCII) {
               pthread_mutex_lock(&transfer_mutex);
               tdata->curfle->startsize += file_countlf(writefd, 0);
               pthread_mutex_unlock(&transfer_mutex);
            }
            g_snprintf(tempstr, sizeof(tempstr), "REST %ld\r\n", tdata->curfle->startsize);
            tempstr[sizeof(tempstr)-1] = '\0';
            queue_log(LOG_SEND, tempstr);
            if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, buf, sizeof(buf), tempstr) != '3') {
               queue_log(LOG_RECV, "%s\n", buf);
               queue_log(LOG_MISC, "Cannot start file transfer at %ld for file %s from %s\n", tdata->curfle->startsize, tdata->curfle->file, tdata->hdata->host);
               break;
            }
            queue_log(LOG_RECV, "%s\n", buf);
         }
         pos = strrchr(tdata->curfle->file, '/');
         if(pos == NULL) pos = tdata->curfle->file;
         else pos++;
         g_snprintf(tempstr, sizeof(tempstr), "RETR %s\r\n", pos);
         tempstr[sizeof(tempstr)-1] = '\0';
         queue_log(LOG_SEND, tempstr);
         if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, buf, sizeof(buf), tempstr) != '1') {
            queue_log(LOG_RECV, "%s\n", buf);
            queue_log(LOG_MISC, "Cannot get file %s\n", tdata->curfle->file);
            break;
         }
         queue_log(LOG_RECV, "%s\n", buf);

         /* The size of the file should be sent in ( ) with the return line
             after the RETR command */
         pos = strchr(buf, '(');

         pthread_mutex_lock(&transfer_mutex);
         if(pos == NULL) tdata->curfle->size = 0;
         else tdata->curfle->size = strtol(pos+1, (char **) NULL, 10);
         tdata->starttime = time(NULL);
         tdata->curtrans = 0;
         pthread_mutex_unlock(&transfer_mutex);
         insert_commas(tdata->curfle->size, ofstr, sizeof(ofstr));

         while(!(tdata->flags & TRANSFER_CANCEL) && (n = read(datafd, &buf, sizeof(buf))) > 0) {
            pthread_mutex_lock(&transfer_mutex);
            tdata->curtrans += n;
            if(tdata->curfle->flags & FILE_RESTART) insert_commas(tdata->curtrans+tdata->curfle->startsize,
               gotstr, sizeof(gotstr));
            else insert_commas(tdata->curtrans, gotstr, sizeof(gotstr));
            kbs = ((float) tdata->curtrans / 1024) / (float) (time(NULL)-tdata->starttime);
            g_snprintf(tdata->progressstr, sizeof(tdata->progressstr), "(%d of %d) Received %s of %s at %.2fkbs",
               num, tdata->hdata->totalfiles, gotstr, ofstr, kbs);
            pthread_mutex_unlock(&transfer_mutex);
            
            if(tdata->hdata->type == FTP_ASCII) { 
               for(i=0, j=0; i<n; i++) if(buf[i] != '\r') buf[j++] = buf[i];
               buf[j] = '\0';
            }
            if(!fwrite(buf, 1, tdata->hdata->type == FTP_ASCII ? j : n, writefd)) break;
         }   
         fclose(writefd);
         close(datafd);
         if(ftp_read_response(0, tdata->wdata, tdata->hdata, NULL, 0) != '2') {
            queue_log(LOG_MISC, "Could not download %s\n", tdata->curfle->file);
         }
         else {
            queue_log(LOG_MISC, "Successfully downloaded %s\n", tdata->curfle->file);
         }
         pthread_mutex_lock(&transfer_mutex);
         tdata->curfle = tdata->curfle->next;
         pthread_mutex_unlock(&transfer_mutex);
      }
      ftp_disconnect(tdata->wdata, tdata->hdata, 0);
   }

   pthread_mutex_lock(&transfer_mutex);
   tdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
   tdata->flags |= TRANSFER_DONE;
   pthread_mutex_unlock(&transfer_mutex);
   pthread_exit(NULL);
}
/*****************************************************************************/
void *ftp_put_files(void *ptr) {
   char tempstr[MAXSTR], buf[8192], gotstr[MAXSTR], ofstr[MAXSTR], *pos, *newbuf;
   struct ftp_transfer_data *tdata;
   int datafd, i, j=0;
   FILE *readfd;
   float kbs;
   ssize_t n, newsize;
   int num;

   num = 0;
   tdata = (struct ftp_transfer_data *) ptr;
   pthread_detach(pthread_self());
   if(ftp_connect(tdata->wdata, tdata->hdata, tdata)) {
      if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, NULL, 0, "TYPE %c\r\n", tdata->hdata->type == FTP_ASCII ? 'A' : 'I') != '2') {
         queue_log(LOG_MISC, "Error: Could not set %s mode\n", tdata->hdata->type == FTP_ASCII ? "ascii" : "binary");
      }
      else while(!(tdata->flags & TRANSFER_CANCEL) && tdata->curfle != NULL) {
         num++;
         if((datafd = ftp_init_data_conn(tdata->wdata, tdata->hdata)) == 0) break;

         if(!(readfd = fopen(tdata->curfle->file, "r"))) {
            queue_log(LOG_MISC, "Error: Cannot open local file %s: %s\n", tdata->curfle->file, g_strerror(errno));
            break;
         }
         if(tdata->curfle->flags & FILE_RESTART) {
            pos = strrchr(tdata->curfle->file, '/');
            if(pos == NULL) pos = tdata->curfle->file;
            else pos++;
            if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, buf, sizeof(buf), "SIZE %s\r\n", pos) != '2') {
               queue_log(LOG_MISC, "Cannot get size of %s on %s\n", pos, tdata->hdata->host);
               break;
            }

            pos = buf;
            pos += 4;
            pthread_mutex_lock(&transfer_mutex);
            tdata->curfle->startsize = strtol(pos, (char **) NULL, 10);
            if(tdata->hdata->type == FTP_ASCII) {
               tdata->curfle->startsize += file_countlf(readfd, tdata->curfle->startsize);
            }
            pthread_mutex_unlock(&transfer_mutex);
            g_snprintf(tempstr, sizeof(tempstr), "REST %ld\r\n", tdata->curfle->startsize);
            tempstr[sizeof(tempstr)-1] = '\0';
            queue_log(LOG_SEND, tempstr);
            if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, buf, sizeof(buf), tempstr) != '3') {
               queue_log(LOG_RECV, "%s\n", buf);
               queue_log(LOG_MISC, "Cannot start file transfer at %ld for file %s from %s\n", tdata->curfle->startsize, tdata->curfle->file, tdata->hdata->host);
               break;
            }
            queue_log(LOG_RECV, "%s\n", buf);
         }
         pos = strrchr(tdata->curfle->file, '/');
         if(pos == NULL) pos = tdata->curfle->file;
         else pos++;
         g_snprintf(tempstr, sizeof(tempstr), "STOR %s\r\n", pos);
         tempstr[sizeof(tempstr)-1] = '\0';
         queue_log(LOG_SEND, tempstr);
         if(ftp_sendcommand(0, tdata->wdata, tdata->hdata, buf, sizeof(buf), tempstr) != '1') {
            queue_log(LOG_RECV, "%s\n", buf);
            queue_log(LOG_MISC, "Cannot put file %s to %s\n", tdata->curfle->file, tdata->hdata->host);
            break;
         }
         queue_log(LOG_RECV, "%s\n", buf);

         pthread_mutex_lock(&transfer_mutex);
         tdata->starttime = time(NULL);
         tdata->curtrans = 0;
         pthread_mutex_unlock(&transfer_mutex);
         insert_commas(tdata->curfle->size, ofstr, sizeof(ofstr));

         while(!(tdata->flags & TRANSFER_CANCEL) && (n = fread(&buf, 1, sizeof(buf), readfd)) > 0) {
            pthread_mutex_lock(&transfer_mutex);
            tdata->curtrans += n;
            if(tdata->curfle->flags & FILE_RESTART) insert_commas(tdata->curtrans+tdata->curfle->startsize,
               gotstr, sizeof(gotstr));
            else insert_commas(tdata->curtrans, gotstr, sizeof(gotstr));
            kbs = ((float) tdata->curtrans / 1024) / (float) (time(NULL)-tdata->starttime);
            g_snprintf(tdata->progressstr, sizeof(tdata->progressstr), "(%d of %d) Sent %s of %s at %.2fkbs",
               num, tdata->hdata->totalfiles, gotstr, ofstr, kbs);
            pthread_mutex_unlock(&transfer_mutex);

            if(tdata->hdata->type == FTP_ASCII) { 
               newsize = 1;
               for(i=0; i<n; i++) {
                  newsize++;
                  if(i > 0 && buf[i] == '\n' && buf[i-1] != '\r') newsize++;
               }
               newbuf = mymalloc(newsize);
               for(i=0,j=0; i<n; i++) {
                  if(i > 0 && buf[i] == '\n' && buf[i-1] != '\r') newbuf[j++] = '\r';
                  newbuf[j++] = buf[i];
               }
               newbuf[newsize-1] = '\0';
            }
            else {
               newbuf = buf;
               newsize = n;
            }

            if(!write(datafd, newbuf, newsize)) {
               if(tdata->hdata->type == FTP_ASCII) free(newbuf);
               queue_log(LOG_MISC, "Cannot write %s to %s\n", tdata->curfle->file, tdata->hdata->host);
               break;
            }
            if(tdata->hdata->type == FTP_ASCII) free(newbuf);
         }   
         fclose(readfd);
         close(datafd);
         if(ftp_read_response(0, tdata->wdata, tdata->hdata, NULL, 0) != '2') {
            queue_log(LOG_MISC, "Could not upload %s to %s\n", tdata->curfle->file, tdata->hdata->host);
         }
         else {
            queue_log(LOG_MISC, "Successfully uploaded %s to %s\n", tdata->curfle->file, tdata->hdata->host);
         }
         pthread_mutex_lock(&transfer_mutex);
         tdata->curfle = tdata->curfle->next;
         pthread_mutex_unlock(&transfer_mutex);
      }
      ftp_disconnect(tdata->wdata, tdata->hdata, 0);
   }

   pthread_mutex_lock(&transfer_mutex);
   tdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
   tdata->flags |= TRANSFER_DONE;
   pthread_mutex_unlock(&transfer_mutex);
   pthread_exit(NULL);
}
/*****************************************************************************/
