/*****************************************************************************/
/*  gftp.c - the main user interface for the ftp program                     */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

/* This are the pixmaps and the masks for some of the built in icons */
GdkPixmap *dotdot_pixmap, *dir_pixmap, *linkdir_pixmap, *linkfile_pixmap,
   *exe_pixmap, *doc_pixmap, *up_pixmap, *down_pixmap;
GdkBitmap *dotdot_mask, *dir_mask, *linkdir_mask, *linkfile_mask, 
   *exe_mask, *doc_mask, *up_mask, *down_mask;

struct pix_ext *registered_exts; /* Registered file extensions */
struct conn_categories *hostcat; /* Our host categories/sites for the connection
                                    manager */

struct ftp_transfer_data *file_transfers = NULL; /* All of our transfers */
pthread_mutex_t transfer_mutex = PTHREAD_MUTEX_INITIALIZER; /* Mutex for 
                                    protecting the file_transfers struct */

struct ftp_log_queue *file_transfer_logs = NULL; /* Waiting log entries from
                                    the threads */
pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER; /* Mutex for protecting
                                    the file_transfer_logs struct */

GtkWidget *logwdw, /* Logging text widget */
          *dlwdw; /* Download list box */

struct ftp_window_data window1, /* Local window data structs */
                       window2; /* Remote window data structs */

/* Several options */
char emailaddr[MAXSTR] = "", /* What to send when we log in as anonymous */
     version[] = "gFTP 1.0"; /* Version string */
int do_one_transfer_at_a_time = 1, /* Only allow one transfer at a time. The other
                                      transfers will be queued */
    start_file_transfers = 1, /* Shall we start the file transfers yet? */
    transfer_in_progress = 0, /* Any transfers transferring files? */
    use_cache = 1; /* Shall we use the cache? */
/*****************************************************************************/
int main(int argc, char *argv[]) {
   GtkWidget *window, *ui;

   gtk_init(&argc, &argv);
   if(argc > 1 && strcmp(argv[1], "--help") == 0) usage();

   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(exitCB), NULL);
   gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(exitCB), NULL);
   gtk_window_set_title(GTK_WINDOW(window), version);
   
   ui = CreateFTPWindows(window);
   gtk_container_add(GTK_CONTAINER(window), ui);
   gtk_widget_show(window);

   ftp_log(LOG_MISC, "%s, Copyright (C) 1998-1999 Brian Masney <", version);
   ftp_log(LOG_RECV, "masneyb@newwave.net");
   ftp_log(LOG_MISC, ">\nIf you have any questions, comments, or suggestions about this program, please\nfeel free to email them to me.\n");
   ftp_log(LOG_MISC, "gFTP comes with ABSOLUTELY NO WARRANTY; for details, see the COPYING file\n");
   ftp_log(LOG_MISC, "This is free software, and you are welcome to redistribute it under certain conditions;\nfor details, see the COPYING file\n");
   init_gftp(argc, argv, window);
                
   add_local_files(&window1); 
   gtk_timeout_add(1000, update_downloads, NULL);

   gtk_main();
   return(0);
}  
/*****************************************************************************/
void init_gftp(int argc, char *argv[], GtkWidget *parent) {
   struct pix_ext *tempext;
   GtkWidget *sort_wid;
   struct utsname unme;
   struct passwd *pw;
   struct hostent *hent;

   window1.local = window2.local = -1;
   if(access(SHARE_DIR, F_OK) == -1) {
      printf("gFTP Error: Cannot find directory %s with the master config files\n", SHARE_DIR);
      printf("Did you do a make install?\n");
      exit(-1);
   }
   read_config_file();
   if(strcmp(emailaddr, "") == 0) {
      /* If there is no email address specified, then we'll just use the
         currentuser@currenthost */
      pw = getpwuid(geteuid());
      uname(&unme);
      if(strchr(unme.nodename, '.')) {
         g_snprintf(emailaddr, sizeof(emailaddr), "%s@%s", pw->pw_name, unme.nodename);
      }
      else {
         hent = gethostbyname(unme.nodename);
         if(hent != NULL) {
            g_snprintf(emailaddr, sizeof(emailaddr), "%s@%s", pw->pw_name, hent->h_name);
         }
         else {
            g_snprintf(emailaddr, sizeof(emailaddr), "%s@%s", pw->pw_name, unme.nodename);
         }
      }
      emailaddr[sizeof(emailaddr)-1] = '\0';
      write_config_file();
   }
                        
   open_xpm("dotdot.xpm", parent, &dotdot_pixmap, &dotdot_mask);
   open_xpm("dir.xpm", parent, &dir_pixmap, &dir_mask);
   open_xpm("linkdir.xpm", parent, &linkdir_pixmap, &linkdir_mask);
   open_xpm("linkfile.xpm", parent, &linkfile_pixmap, &linkfile_mask);
   open_xpm("exe.xpm", parent, &exe_pixmap, &exe_mask);
   open_xpm("doc.xpm", parent, &doc_pixmap, &doc_mask);
   open_xpm("up.xpm", parent, &up_pixmap, &up_mask);
   open_xpm("down.xpm", parent, &down_pixmap, &down_mask);
   
   if(window1.sortasds) sort_wid = gtk_pixmap_new(down_pixmap, down_mask);
   else sort_wid = gtk_pixmap_new(up_pixmap, up_mask);
   gtk_widget_show(sort_wid);
   gtk_clist_set_column_widget(GTK_CLIST(window1.listbox), 0, sort_wid);

   if(window2.sortasds) sort_wid = gtk_pixmap_new(down_pixmap, down_mask);
   else sort_wid = gtk_pixmap_new(up_pixmap, up_mask);
   gtk_widget_show(sort_wid);
   gtk_clist_set_column_widget(GTK_CLIST(window2.listbox), 0, sort_wid);

   tempext = registered_exts;
   while(tempext != NULL) {
      open_xpm(tempext->filename, parent, &tempext->pixmap, &tempext->mask);
      tempext = tempext->next;
   }
   if(argc > 2) usage();
   else if(argc == 2) {
      if(parse_ftp_url(window2.host, argv[1])) {
         fix_display();
         if(ftp_connect(&window2, window2.host, NULL)) {
            ftp_list_files(&window2, 1);
         }
      }
      else usage();
   }
}
/*****************************************************************************/
int parse_ftp_url(struct ftp_host_data *hdata, char *url) {
   char tempstr[MAXSTR], *pos, *endpos;
   int len;

   hdata->port = 0;
   hdata->host[0] = hdata->dir[0] = '\0';
   strncpy(hdata->user, ANON_LOGIN, sizeof(hdata->user));
   hdata->user[sizeof(hdata->user)-1] = '\0';
   strncpy(hdata->passwd, emailaddr, sizeof(hdata->passwd));
   hdata->passwd[sizeof(hdata->passwd)-1] = '\0';
   pos = url;
   if(strncmp(pos, "ftp://", 6) == 0) pos += 6;
   if(strchr(pos, '@') != NULL) {
      /* A user/password was entered */
      if((endpos = strchr(pos, ':')) == NULL) return(0);
      len = endpos - pos + 1 > sizeof(hdata->user) ? sizeof(hdata->user) : endpos - pos + 1;
      strncpy(hdata->user, pos, len);
      hdata->user[len-1] = 0;

      pos = endpos+1;
      if((endpos = strchr(pos, '@')) == NULL) return(0);
      if(strchr(endpos+1, '@') != NULL) endpos = strchr(endpos+1, '@');
      len = endpos - pos + 1 > sizeof(hdata->passwd) ? sizeof(hdata->passwd) : endpos - pos + 1;
      strncpy(hdata->passwd, pos, len);
      hdata->passwd[len-1] = 0;
      pos = endpos+1;
   }
   /* Now get the hostname and an optional port and optional directory */
   if((endpos = strchr(pos, ':')) != NULL) {
      len = endpos - pos + 1 > sizeof(hdata->host) ? sizeof(hdata->host) : endpos - pos + 1;
   }
   else if((endpos = strchr(pos, '/')) != NULL) {
      len = endpos - pos + 1 > sizeof(hdata->host) ? sizeof(hdata->host) : endpos - pos + 1;
   }
   else len = sizeof(hdata->host);
   strncpy(hdata->host, pos, len);
   hdata->host[len-1] = 0;
   if((endpos = strchr(pos, ':')) != NULL) {
      /* A port was entered */
      pos = endpos + 1;
      if((endpos = strchr(pos, '/')) != NULL) {
         len = endpos - pos + 1 > sizeof(tempstr) ? sizeof(tempstr) : endpos - pos + 1;
      }
      g_snprintf(tempstr, len, pos);
      tempstr[len-1] = 0;
      hdata->port = strtol(pos, (char **) NULL, 10);
   }
   if((endpos = strchr(pos, '/')) != NULL) {
      strncpy(hdata->dir, endpos, sizeof(hdata->dir));
      hdata->dir[sizeof(hdata->dir)-1] = '\0';
   }
   return(1);
}
/*****************************************************************************/
void open_xpm(char *filename, GtkWidget *parent, GdkPixmap **pixmap, GdkBitmap **mask) {
   char tempstr[MAXSTR], temp1str[MAXSTR];
   GtkStyle *style;

   style = gtk_widget_get_style(parent);
   g_snprintf(temp1str, sizeof(temp1str), "%s/%s", SHARE_DIR, filename);
   temp1str[sizeof(temp1str)-1] = '\0';
   expand_path(temp1str, tempstr, sizeof(tempstr));
   if(access(tempstr, F_OK) != 0) {
      g_snprintf(temp1str, sizeof(temp1str), "%s/%s", BASE_CONF_DIR, filename);
      temp1str[sizeof(temp1str)-1] = '\0';
      expand_path(temp1str, tempstr, sizeof(tempstr));
      if(access(tempstr, F_OK) != 0) {
         printf("gFTP Error: Cannot find file %s in %s or %s\n", filename, SHARE_DIR, BASE_CONF_DIR);
         exit(-1);
      }
   }
   *pixmap = gdk_pixmap_create_from_xpm(parent->window, mask, &style->bg[GTK_STATE_NORMAL], tempstr);
   if(*pixmap == NULL) {
      printf("gFTP Error: Error opening file %s\n", tempstr);
      exit(-1);
   }
}
/*****************************************************************************/
void usage(void) {
   printf("usage: gftp [[ftp://][user:pass@]ftp-site[:port][/directory]]\n");
   exit(0);
}
/*****************************************************************************/
GtkWidget *CreateFTPWindows(GtkWidget *ui) {
   GtkWidget *win1frame = NULL, *win2frame = NULL, *putbutton, *retrbutton;
   GtkWidget *box, *dlbox, *winpane, *dlpane, *logpane, *vscrollbar, *logtable;
   GtkWidget *mainvbox, *scroll_list, *menuitem;
   GtkItemFactory *factory;
   GtkAccelGroup *accel_group;
   char *dltitles[3] = {"Filename", "Progress", "Hostname"};
   GtkItemFactoryEntry menu_items[] = {
      {"/_FTP",			NULL,	0,		0,	"<Branch>"},
      {"/FTP/tearoff",		NULL,	0,		0,	"<Tearoff>"},
      {"/FTP/Ascii", 		NULL, 	change_ascii,	0,	"<RadioItem>"},
      {"/FTP/Binary", 		NULL, 	change_binary,	0,	"/FTP/Ascii"},
      {"/FTP/sep", 		NULL, 	0, 		0,	"<Separator>"},
      {"/FTP/Use Cache",	NULL,	change_use_cache, 0,	"<CheckItem>"},
      {"/FTP/Start Transfers",	NULL,	change_st_trans, 0,	"<CheckItem>"},
      {"/FTP/Do only one transfer", NULL, change_one_trans, 0,	"<CheckItem>"},
      {"/FTP/sep", 		NULL, 	0, 		0,	"<Separator>"},
      {"/FTP/_Quit", 		"<control>Q", 	exitCB,	0},
      {"/_Local",		NULL,	0,		0,	"<Branch>"},
      {"/Local/tearoff",	NULL,	0,		0,	"<Tearoff>"},
      {"/Local/Select All", 	NULL, 	selectall, 	0},
      {"/Local/Deselect All", 	NULL, 	deselectall, 	0},
      {"/Local/sep",		NULL,	0,		0,	"<Separator>"},
      {"/Local/Change Directory", NULL,	chdirfunc, 	0},
      {"/Local/Make Directory...", NULL, mkdir_dialog, 	0},
      {"/Local/Rename...", 	NULL, 	rename_dialog, 	0},
      {"/Local/Delete...", 	NULL, 	delete_dialog, 	0},
      {"/Local/View...",	NULL,	view_dialog,	0},
      {"/Local/Refresh", 	NULL, 	refresh, 	0},
      {"/_Remote",		NULL,	0,		0,	"<Branch>"},
      {"/Remote/tearoff",	NULL,	0,		0,	"<Tearoff>"},
      {"/Remote/_Connect...", 	"<control>C", 	connect_dialog, 0},
      {"/Remote/Open URL...",	NULL,	openurl_dialog,	0},
      {"/Remote/_Disconnect", 	"<control>D", 	disconnect, 	0},
      {"/Remote/sep",		NULL,	0,		0,	"<Separator>"},
      {"/Remote/Select All", 	NULL, 	selectall, 	0},
      {"/Remote/Deselect All", 	NULL, 	deselectall, 	0},
      {"/Remote/sep",		NULL,	0,		0,	"<Separator>"},
      {"/Remote/Change Directory", NULL, chdirfunc, 	0},
      {"/Remote/Make Directory...", NULL, mkdir_dialog,	0},
      {"/Remote/Rename...", 	NULL, 	rename_dialog, 	0},
      {"/Remote/Delete...", 	NULL, 	delete_dialog, 	0},
      {"/Remote/View...",	NULL,	view_dialog,	0},
      {"/Remote/Refresh", 	NULL, 	refresh, 	0},
      {"/_Transfers",		NULL,	0,		0,	"<Branch>"},
      {"/Transfers/tearoff",	NULL,	0,		0,	"<Tearoff>"},
      {"/Transfers/Stop Transfer", NULL, stop_transfer,	0},
      {"/Transfers/sep",	NULL,	0,		0,	"<Separator>"},
      {"/Transfers/Retrieve Files", NULL,  retrmenu,	0},
      {"/Transfers/Put Files",	NULL,	putmenu,	0},
      {"/L_ogging",		NULL,	0,		0,	"<Branch>"},
      {"/Logging/tearoff",	NULL,	0,		0,	"<Tearoff>"},
      {"/Logging/Clear", 	NULL, 	clearlog, 	0},
      {"/Logging/Save log...", 	NULL, 	savelog, 	0},
      {"/Tool_s",		NULL,	0,		0,	"<Branch>"},
      {"/Tools/tearoff",	NULL,	0,		0,	"<Tearoff>"},
      {"/Tools/Compare Windows", NULL, 	compare_windows,  0}};
      
   mainvbox = gtk_vbox_new(FALSE, 0);
   gtk_widget_show(mainvbox);

   accel_group = gtk_accel_group_new();
   factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);

   /* FTP Menu */
   gtk_item_factory_create_items(factory, 10, menu_items, &window2);
   /* Local Menu */
   gtk_item_factory_create_items(factory, 11, menu_items + 10, &window1);
   /* Remote Menu */
   gtk_item_factory_create_items(factory, 15, menu_items + 21, &window2);
   /* Transfers Menu */
   gtk_item_factory_create_items(factory, 6, menu_items + 36, NULL);
   /* Logging Menu */
   gtk_item_factory_create_items(factory, 4, menu_items + 42, NULL);
   /* Tools Menu */
   gtk_item_factory_create_items(factory, 3, menu_items + 46, NULL);
   
   gtk_accel_group_attach(accel_group, GTK_OBJECT(ui));
   gtk_box_pack_start(GTK_BOX(mainvbox), factory->widget, FALSE, FALSE, FALSE);
   gtk_widget_show(factory->widget);

   menuitem = gtk_item_factory_get_widget(factory, menu_items[2].path);
   gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem), FALSE);

   menuitem = gtk_item_factory_get_widget(factory, menu_items[3].path);
   gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem), TRUE);

   menuitem = gtk_item_factory_get_widget(factory, menu_items[5].path);
   gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem), use_cache);

   menuitem = gtk_item_factory_get_widget(factory, menu_items[6].path);
   gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem), start_file_transfers);

   menuitem = gtk_item_factory_get_widget(factory, menu_items[7].path);
   gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem), do_one_transfer_at_a_time);

   winpane = gtk_hpaned_new();
   box = gtk_hbox_new(FALSE, 0);
   win1frame = CreateFTPWindow(win1frame, &window1);
   gtk_box_pack_start(GTK_BOX(box), win1frame, TRUE, TRUE, TRUE);
   gtk_widget_show(win1frame);

   dlbox = gtk_vbox_new(FALSE, 0);
   gtk_container_border_width(GTK_CONTAINER(dlbox), 5);

   putbutton = gtk_button_new_with_label("->");
   gtk_box_pack_start(GTK_BOX(dlbox), putbutton, TRUE, FALSE, FALSE);
   gtk_signal_connect(GTK_OBJECT(putbutton), "clicked", GTK_SIGNAL_FUNC(putCB), (gpointer) NULL);
   gtk_widget_show(putbutton);

   retrbutton = gtk_button_new_with_label("<-");
   gtk_box_pack_start(GTK_BOX(dlbox), retrbutton, TRUE, FALSE, FALSE);
   gtk_signal_connect(GTK_OBJECT(retrbutton), "clicked", GTK_SIGNAL_FUNC(retrCB), (gpointer) NULL);
   gtk_widget_show(retrbutton);

   gtk_box_pack_start(GTK_BOX(box), dlbox, FALSE, FALSE, FALSE);
   gtk_widget_show(dlbox);
   gtk_widget_show(box);
   gtk_paned_add1(GTK_PANED(winpane), box);
   
   win2frame = CreateFTPWindow(win2frame, &window2);
   gtk_paned_add2(GTK_PANED(winpane), win2frame);
   gtk_widget_show(win2frame);
   gtk_widget_show(winpane);

   dlpane = gtk_vpaned_new();
   gtk_paned_add1(GTK_PANED(dlpane), winpane);

   scroll_list = gtk_scrolled_window_new(NULL, NULL);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll_list),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   dlwdw = gtk_clist_new_with_titles(3, dltitles);
   gtk_container_add(GTK_CONTAINER(scroll_list), dlwdw);
   gtk_widget_set_usize(dlwdw, 1, 50);
   gtk_clist_set_column_width(GTK_CLIST(dlwdw), 0, 100);
   gtk_clist_set_column_width(GTK_CLIST(dlwdw), 1, 250);
   gtk_clist_set_column_width(GTK_CLIST(dlwdw), 2, 100);
   gtk_container_border_width(GTK_CONTAINER(dlwdw), 5);
   gtk_signal_connect(GTK_OBJECT(dlwdw), "select_row", GTK_SIGNAL_FUNC(selectdl), NULL);
   gtk_signal_connect(GTK_OBJECT(dlwdw), "unselect_row", GTK_SIGNAL_FUNC(unselectdl), NULL);
   gtk_paned_add2(GTK_PANED(dlpane), scroll_list);
   gtk_widget_show(dlwdw);
   gtk_widget_show(scroll_list);
   gtk_widget_show(dlpane);

   logpane = gtk_vpaned_new();
   gtk_paned_add1(GTK_PANED(logpane), dlpane);

   logtable = gtk_table_new(1, 2, FALSE);
   logwdw = gtk_text_new(NULL, NULL);
   gtk_widget_set_usize(logwdw, 1, 105);
   gtk_text_set_editable(GTK_TEXT(logwdw), FALSE);
   gtk_table_attach(GTK_TABLE(logtable), logwdw, 0, 1, 0, 1,
      GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 0);
   gtk_widget_show(logwdw);

   vscrollbar = gtk_vscrollbar_new (GTK_TEXT(logwdw)->vadj);
   gtk_table_attach(GTK_TABLE(logtable), vscrollbar, 1, 2, 0, 1,
      GTK_FILL, GTK_EXPAND | GTK_FILL | GTK_SHRINK, 0, 0);
   gtk_widget_show(vscrollbar);
   gtk_widget_show(logtable);
   
   gtk_paned_add2(GTK_PANED(logpane), logtable);
   gtk_widget_show(logpane);      
   gtk_box_pack_start(GTK_BOX(mainvbox), logpane, TRUE, TRUE, TRUE);

   return(mainvbox);
}   
/*****************************************************************************/
GtkWidget *CreateFTPWindow(GtkWidget *parent, struct ftp_window_data *wdata) {
   char *titles[7] = {"", "Filename", "Size", "User", "Group", "Date", "Attribs"};
   GtkTargetEntry possible_types[] = {
      {"STRING",			0,	0},
      {"text/plain", 			0, 	0},
      {"application/x-rootwin-drop", 	0, 	1}};
   GtkWidget *box, *scroll_list;

   wdata->host = mymalloc(sizeof(struct ftp_host_data));
   wdata->host->files = wdata->host->last = NULL;
   wdata->host->type = FTP_BINARY;
   memset(wdata->host->recvdata.recvdata, 0, MAXSTR);
   wdata->host->recvdata.pos = wdata->host->recvdata.recvdata;
         
   wdata->sortcol = 1;
   wdata->sortasds = 1;
   wdata->totalitems = wdata->numselected = 0;
   wdata->items = NULL;

   parent = gtk_frame_new(NULL);
   gtk_container_border_width(GTK_CONTAINER(parent), 5);
   gtk_widget_set_usize(parent, 255, 275);
   
   box = gtk_vbox_new(FALSE, 0);
   gtk_container_border_width(GTK_CONTAINER(box), 5);
   gtk_container_add(GTK_CONTAINER(parent), box);
   
   wdata->diredit = gtk_entry_new_with_max_length(MAXSTR);
   gtk_box_pack_start(GTK_BOX(box), wdata->diredit, FALSE, FALSE, FALSE);
   gtk_signal_connect(GTK_OBJECT(wdata->diredit), "activate", GTK_SIGNAL_FUNC(chdiredit), (gpointer) wdata);
   gtk_widget_show(wdata->diredit);
   
   wdata->hoststxt = gtk_label_new("Not connected");
   gtk_misc_set_alignment(GTK_MISC(wdata->hoststxt), 0, 0);
   gtk_box_pack_start(GTK_BOX(box), wdata->hoststxt, FALSE, FALSE, FALSE);
   gtk_widget_show(wdata->hoststxt);

   scroll_list = gtk_scrolled_window_new(NULL, NULL);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll_list),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   wdata->listbox = gtk_clist_new_with_titles(7, titles);
   gtk_container_add(GTK_CONTAINER(scroll_list), wdata->listbox);
   gtk_drag_source_set(wdata->listbox, GDK_BUTTON2_MASK, possible_types, 3,
      GDK_ACTION_COPY | GDK_ACTION_MOVE);
   gtk_drag_dest_set(wdata->listbox, GTK_DEST_DEFAULT_ALL, possible_types, 2,
      GDK_ACTION_COPY | GDK_ACTION_MOVE);
   gtk_clist_set_selection_mode(GTK_CLIST(wdata->listbox), GTK_SELECTION_EXTENDED);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 0, 16);
   gtk_clist_set_column_justification(GTK_CLIST(wdata->listbox), 0, GTK_JUSTIFY_CENTER);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 1, 100);
   gtk_clist_set_column_justification(GTK_CLIST(wdata->listbox), 2, GTK_JUSTIFY_RIGHT);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 2, 85);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 3, 65);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 4, 65);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 5, 85);
   gtk_clist_set_column_width(GTK_CLIST(wdata->listbox), 6, 65);
   gtk_box_pack_start(GTK_BOX(box), scroll_list, TRUE, TRUE, TRUE);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "click_column", GTK_SIGNAL_FUNC(sortrows), (gpointer) wdata);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "select_row", GTK_SIGNAL_FUNC(selectrow), (gpointer) wdata);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "unselect_row", GTK_SIGNAL_FUNC(unselectrow), (gpointer) wdata);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "button_press_event", GTK_SIGNAL_FUNC(list_dblclick), (gpointer) wdata);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "drag_data_get", GTK_SIGNAL_FUNC(listbox_drag), (gpointer) wdata);
   gtk_signal_connect(GTK_OBJECT(wdata->listbox), "drag_data_received", GTK_SIGNAL_FUNC(listbox_get_drag_data), (gpointer) wdata);
   gtk_widget_show(wdata->listbox);
   gtk_widget_show(scroll_list);

   gtk_widget_show(box);
   gtk_widget_show(parent);

   return(parent);
}
/*****************************************************************************/
void listbox_drag(GtkWidget *widget, GdkDragContext *context, GtkSelectionData *selection_data, guint info, guint32 clk_time, gpointer data) {
   struct ftp_window_data *wdata;
   struct ftp_file_data *tempfle;
   char *str = NULL;
   size_t totlen = 0, oldlen;
   char tempstr[MAXSTR];
   int i;
   
   wdata = (struct ftp_window_data *) data;
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Drag-N-Drop: Not connected to a remote site\n");
      return;
   }
   if(wdata->numselected == 0) {
      ftp_log(LOG_MISC, "Drag-N-Drop: You must have at least one item selected\n");
      return;
   }

   tempfle = wdata->host->files;
   for(i=0; i<wdata->totalitems; i++) {
      if(wdata->items[i]) {
         if(totlen != 0) totlen--;
         oldlen = totlen;
         if(wdata->local == 1) {
            g_snprintf(tempstr, sizeof(tempstr), "%s/%s\n", wdata->host->dir, tempfle->file);
         }
         else {
            g_snprintf(tempstr, sizeof(tempstr), "ftp://%s:%s@%s:%ld%s/%s\n", 
               wdata->host->user, wdata->host->passwd, wdata->host->host, 
               wdata->host->port == 0 ? 21 : wdata->host->port,
               wdata->host->dir, tempfle->file);
         }
         tempstr[sizeof(tempstr)-1] = '\0';
         remove_double_slashes(tempstr+6);
         totlen += strlen(tempstr)+1;
         str = myrealloc(str, totlen);
         strcpy(str+oldlen, tempstr);
      }
      tempfle = tempfle->next;
   }
   if(totlen > 1) str[totlen-2] = '\0';
   gtk_selection_data_set(selection_data, selection_data->target, 8,
      str, strlen(str));
   free(str);
}
/*****************************************************************************/
void listbox_get_drag_data(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *selection_data, guint info, guint32 clk_time, struct ftp_window_data *wdata) {
   struct ftp_file_data *newfle;
   struct ftp_transfer_data *tdata, *newtdata, *temptdata, *new_file_transfers;
   char *newpos, *oldpos, *pos;
   char tempstr[MAXSTR];
   struct stat st;
   int len, ret, finish_drag;
   
   finish_drag = 0;
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Drag-N-Drop: Not connected to a remote site\n");
      return;
   }
   new_file_transfers = NULL;
   if((selection_data->length >= 0) && (selection_data->format == 8)) {
      if(!wdata->local) {
         tdata = create_new_tdata();
         *tdata->hdata = *(&window2)->host;
         tdata->hdata->files = tdata->hdata->last = NULL;
         tdata->hdata->totalfiles = 0;
         if(wdata->local) {
            tdata->flags |= TRANSFER_DIRECTION;
            tdata->wdata = &window2;
            tdata->otherwdata = &window1;
         }
         else {
            tdata->wdata = &window1;
            tdata->otherwdata = &window2;
         }
         tdata->curfle = NULL;
         tdata->next = new_file_transfers;
         new_file_transfers = tdata;
      }
      else tdata = NULL;
      oldpos = (char *) selection_data->data;
      while((newpos = strchr(oldpos, '\n')) || (newpos = strchr(oldpos, '\0'))) {
         len = newpos - oldpos + 1 > sizeof(tempstr) ? sizeof(tempstr) : newpos - oldpos + 1;
         strncpy(tempstr, oldpos, len);
         tempstr[len-1] = '\0';
         ret = strncmp(oldpos, "ftp://", 6);
         if(!wdata->local && ret == 0) {
            ftp_log(LOG_MISC, "Drag-N-Drop: Ignoring file %s: Cannot drag a ftp site to a ftp site\n", tempstr);
            if(*newpos == '\0') break;
            oldpos = newpos + 1;
            continue;
         }
         else if(wdata->local && ret != 0) {
            ftp_log(LOG_MISC, "Drag-N-Drop: Ignoring file %s: Cannot drag a local file to the local window\n", tempstr);
            if(*newpos == '\0') break;
            oldpos = newpos + 1;
            continue;
         }
         else {
            newfle = mymalloc(sizeof(struct ftp_file_data));
            if(wdata->local) {
               newtdata = create_new_tdata();
               newtdata->hdata->files = newtdata->hdata->last = NULL;
               newtdata->hdata->totalfiles = 0;
               if(wdata->local) {
                  newtdata->flags |= TRANSFER_DIRECTION;
                  newtdata->wdata = &window2;
                  newtdata->otherwdata = &window1;
               }
               else {
                  newtdata->wdata = &window1;
                  newtdata->otherwdata = &window2;
               }
               newtdata->curfle = NULL;
               newtdata->hdata->type = window2.host->type;
               if(!parse_ftp_url(newtdata->hdata, tempstr)) {
                  ftp_log(LOG_MISC, "Drag-N-Drop: Ignoring url %s: Not a valid url\n", tempstr);
                  if(*newpos == '\0') break;
                  oldpos = newpos + 1;
                  free(newtdata);
                  continue;
               }
               
               pos = strrchr(newtdata->hdata->dir, '/');
               if(pos == NULL) pos = newtdata->hdata->dir;
               g_snprintf(newfle->file, sizeof(newfle->file), "%s/%s", window1.host->dir, pos);
               newfle->file[sizeof(newfle->file)-1] = '\0';
               remove_double_slashes(newfle->file);
               newfle->size = 0;
               *pos = '\0';
               temptdata = new_file_transfers;
               while(temptdata != NULL) {
                  if(compare_hdata_structs(temptdata->hdata, newtdata->hdata)) {
                     tdata = temptdata;
                     break;
                  }
                  temptdata = temptdata->next;
               }
               if(temptdata == NULL) {
                  newtdata->next = new_file_transfers;
                  new_file_transfers = newtdata;
                  tdata = newtdata;
               }
               else {
                  tdata = temptdata;
                  free(newtdata);
               }
            }
            else {
               stat(tempstr, &st);
               strncpy(newfle->file, tempstr, sizeof(newfle->file));
               newfle->size = st.st_size;
            }
            newfle->next = NULL;
            if(tdata->hdata->last == NULL) tdata->hdata->files = newfle;
            else tdata->hdata->last->next = newfle;
            tdata->hdata->last = newfle;
            tdata->hdata->totalfiles++;
            if(tdata->curfle == NULL) tdata->curfle = newfle;
            finish_drag = 1;
         }
         if(*newpos == '\0') break;
         oldpos = newpos + 1;
      }
   }
   if(finish_drag) {
      temptdata = new_file_transfers;
      while(temptdata != NULL) {
         dotrans(temptdata);
         temptdata = temptdata->next;
      }

      pthread_mutex_lock(&transfer_mutex);
      temptdata = file_transfers;
      if(temptdata == NULL) file_transfers = new_file_transfers;
      else {
         while(temptdata->next != NULL) temptdata = temptdata->next;
         temptdata->next = new_file_transfers;
      }
      pthread_mutex_unlock(&transfer_mutex);
   }
   gtk_drag_finish(context, finish_drag, FALSE, clk_time);
}
/*****************************************************************************/
void sortrows(GtkCList *clist, gint column, gpointer data) {
   struct ftp_file_data *newfle, *tempfle, *prevfle, *curfle;
   struct ftp_window_data *wdata;
   GtkWidget *sort_wid;

   wdata = (struct ftp_window_data *) data;
   if(!wdata->host->files) return;
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Sort: Not connected to a remote site\n");
      return;
   }
   gtk_label_set(GTK_LABEL(wdata->hoststxt), "Sorting...");
   fix_display();
   newfle = wdata->host->files;
   if(column == 0) {
      wdata->sortasds = !wdata->sortasds;
      column = wdata->sortcol;
      if(wdata->sortasds) sort_wid = gtk_pixmap_new(down_pixmap, down_mask);
      else sort_wid = gtk_pixmap_new(up_pixmap, up_mask);
      gtk_widget_show(sort_wid);
      gtk_clist_set_column_widget(clist, 0, sort_wid);
   }
   else wdata->sortcol = column;
   wdata->host->files = wdata->host->files->next;
   newfle->next = NULL;
   while(wdata->host->files != NULL) {
      curfle = wdata->host->files;
      wdata->host->files = wdata->host->files->next;
      tempfle = prevfle = newfle;
      if(column == 1) {
         while(tempfle != NULL && (wdata->sortasds ?
            strcmp(tempfle->file, curfle->file) <= 0 :
            strcmp(tempfle->file, curfle->file) >= 0)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      else if(column == 2) {
         while(tempfle != NULL && (wdata->sortasds ? 
            tempfle->size <= curfle->size : 
            tempfle->size >= curfle->size)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      else if(column == 3) {
         while(tempfle != NULL && (wdata->sortasds ?
            strcmp(tempfle->user, curfle->user) <= 0 :
            strcmp(tempfle->user, curfle->user) >= 0)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      else if(column == 4) {
         while(tempfle != NULL && (wdata->sortasds ?
            strcmp(tempfle->group, curfle->group) <= 0 :
            strcmp(tempfle->group, curfle->group) >= 0)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      else if(column == 5) {
         while(tempfle != NULL && (wdata->sortasds ? 
            strcmp(tempfle->sort_time, curfle->sort_time) <= 0 :
            strcmp(tempfle->sort_time, curfle->sort_time) >= 0)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      else if(column == 6) {
         while(tempfle != NULL && (wdata->sortasds ?
            strcmp(tempfle->attribs, curfle->attribs) <= 0 :
            strcmp(tempfle->attribs, curfle->attribs) >= 0)) {

            prevfle = tempfle;
            tempfle = tempfle->next;
         }
      }
      if(tempfle == NULL) {
         prevfle->next = curfle;
         curfle->next = NULL;
      }
      else if(tempfle == prevfle) {
         curfle->next = newfle;
         newfle = curfle;
      }
      else {            
         curfle->next = prevfle->next;
         prevfle->next = curfle;
      }
   }
   wdata->host->files = newfle;
   gtk_clist_freeze(clist);
   deselectall(wdata);
   gtk_clist_clear(clist);
   tempfle = newfle;
   while(tempfle != NULL) {
      add_file_listbox(wdata, tempfle);
      tempfle = tempfle->next;
   }
   gtk_clist_thaw(clist);
   update_ftp_info(wdata);
}
/*****************************************************************************/
void delete_ftp_file_info(struct ftp_window_data *wdata) {
   struct ftp_file_data *fle;
      
   gtk_clist_clear(GTK_CLIST(wdata->listbox));
   fle = wdata->host->files;
   while(fle != NULL) {
      wdata->host->files = wdata->host->files->next;
      free(fle);
      fle = wdata->host->files;
   }
   wdata->host->last = NULL;
}
/*****************************************************************************/
void queue_log(int type, char *format, ...) {
   struct ftp_log_queue *newlog, *templog;
   char tempstr[MAXSTR];
   va_list argp;
   
   newlog = mymalloc(sizeof(struct ftp_log_queue));
   newlog->type = type;
   va_start(argp, format);
   g_vsnprintf(tempstr, sizeof(tempstr), format, argp);
   tempstr[sizeof(tempstr)-1] = '\0';
   strncpy(newlog->msg, tempstr, sizeof(newlog->msg));
   newlog->next = NULL;
   pthread_mutex_lock(&log_mutex);
   if(file_transfer_logs == NULL) file_transfer_logs = newlog;
   else {
      templog = file_transfer_logs;
      while(templog->next != NULL) templog = templog->next;
      templog->next = newlog;
   }
   pthread_mutex_unlock(&log_mutex);
}
/*****************************************************************************/
void ftp_log(int type, char *format, ...) {
   char tempstr[MAXSTR];
   GdkColormap *cmap;
   GdkColor fore;
   va_list argp;
   guint pos;
   int upd;
   
   upd = GTK_TEXT(logwdw)->vadj->upper - GTK_TEXT(logwdw)->vadj->page_size == GTK_TEXT(logwdw)->vadj->value;
   cmap = gdk_colormap_get_system();
   fore.red = fore.green = fore.blue = 0;
   if(type == LOG_SEND) fore.green = 0x8600;
   else if(type == LOG_RECV) fore.blue = 0xffff;
   else fore.red = 0xffff;
   if(!gdk_color_alloc(cmap, &fore)) g_error("Could not allocate color for log window");
   gtk_text_freeze(GTK_TEXT(logwdw));
   pos = gtk_text_get_length(GTK_TEXT(logwdw));

   va_start(argp, format);
   g_vsnprintf(tempstr, sizeof(tempstr), format, argp);
   tempstr[sizeof(tempstr)-1] = '\0';
   gtk_text_insert(GTK_TEXT(logwdw), NULL, &fore, NULL, tempstr, -1);
   
   gtk_text_set_point(GTK_TEXT(logwdw), pos+strlen(tempstr));
   gtk_text_thaw(GTK_TEXT(logwdw));
   if(upd) {
      gtk_adjustment_set_value(GTK_TEXT(logwdw)->vadj, 
         GTK_TEXT(logwdw)->vadj->upper - GTK_TEXT(logwdw)->vadj->page_size);
   }
   fix_display();
}
/*****************************************************************************/
void update_ftp_info(struct ftp_window_data *wdata) {
   char tempstr[MAXSTR];
   
   if(wdata->local == 1) {
      gtk_label_set(GTK_LABEL(wdata->hoststxt), "Local");
      gtk_entry_set_text(GTK_ENTRY(wdata->diredit), wdata->host->dir);
   }
   else if(wdata->local == 0) {
      g_snprintf(tempstr, sizeof(tempstr), "%s%s", wdata->host->host, wdata->cached ? " (Cached)" : "");
      tempstr[sizeof(tempstr)-1] = '\0';
      gtk_label_set(GTK_LABEL(wdata->hoststxt), tempstr);
      gtk_entry_set_text(GTK_ENTRY(wdata->diredit), wdata->host->dir);
   }
   else {
      gtk_label_set(GTK_LABEL(wdata->hoststxt), "Not connected");
      gtk_entry_set_text(GTK_ENTRY(wdata->diredit), "");
   }
   fix_display();
}
/*****************************************************************************/
void change_use_cache(gpointer data, int menuitem, GtkWidget *checkmenu) {
   use_cache = GTK_CHECK_MENU_ITEM(checkmenu)->active;
}
/*****************************************************************************/
void change_st_trans(gpointer data, int menuitem, GtkWidget *checkmenu) {
   start_file_transfers = GTK_CHECK_MENU_ITEM(checkmenu)->active;
}
/*****************************************************************************/
void change_one_trans(gpointer data, int menuitem, GtkWidget *checkmenu) {
   do_one_transfer_at_a_time = GTK_CHECK_MENU_ITEM(checkmenu)->active;
}
/*****************************************************************************/
void change_ascii(struct ftp_window_data *wdata) {
   if(wdata->host) wdata->host->type = FTP_ASCII;
}
/*****************************************************************************/
void change_binary(struct ftp_window_data *wdata) {
   if(wdata->host) wdata->host->type = FTP_BINARY;
}
/*****************************************************************************/
void selectrow(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data) {
   struct ftp_window_data *wdata;
   
   wdata = (struct ftp_window_data *) data;
   if(!wdata->items[row]) wdata->numselected++;
   wdata->items[row] = 1;
}
/*****************************************************************************/
void unselectrow(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data) {
   struct ftp_window_data *wdata;
   
   wdata = (struct ftp_window_data *) data;
   if(wdata->items[row]) wdata->numselected--;
   wdata->items[row] = 0;
}
/*****************************************************************************/
void selectall(struct ftp_window_data *wdata) {
   struct ftp_file_data *tempfle;
   int i = 0;
   
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Select All: Not connected to a remote site\n");
      return;
   }
   tempfle = wdata->host->files;
   while(tempfle != NULL) {
      gtk_clist_select_row(GTK_CLIST(wdata->listbox), i++, 0);
      tempfle = tempfle->next;
   }
}
/*****************************************************************************/
void deselectall(struct ftp_window_data *wdata) {
   struct ftp_file_data *tempfle;
   int i = 0;
   
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Deselect All: Not connected to a remote site\n");
      return;
   }
   tempfle = wdata->host->files;
   while(tempfle != NULL) {
      gtk_clist_unselect_row(GTK_CLIST(wdata->listbox), i++, 0);
      tempfle = tempfle->next;
   }
}
/*****************************************************************************/
void list_dblclick(GtkWidget *widget, GdkEventButton *event, gpointer data) {
   struct ftp_window_data *wdata;
   struct ftp_file_data *tempfle;
   int pos, i;
   
   wdata = (struct ftp_window_data *) data;
   if(wdata->local == -1 || wdata->numselected != 1) return;
   if(event->type == GDK_2BUTTON_PRESS) {
      for(i=0; i<wdata->totalitems; i++) if(wdata->items[i]) break;
      pos = i;
      tempfle = wdata->host->files;
      for(i=0; i<pos; i++) tempfle = tempfle->next;
      if(tempfle->flags & FILE_ISDIR) chdirfunc(data);
      else view_dialog(data);
   }
}
/*****************************************************************************/
void chdirfunc(struct ftp_window_data *wdata) {
   struct ftp_file_data *tempfle;
   int pos, i;
   
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Chdir: Not connected to a remote site\n");
      return;
   }
   else if(wdata->numselected != 1) {
      ftp_log(LOG_MISC, "Chdir: You must only have one item selected\n");
      return;
   }
   for(i=0; i<wdata->totalitems; i++) if(wdata->items[i]) break;
   pos = i;
   tempfle = wdata->host->files;
   for(i=0; i<pos; i++) tempfle = tempfle->next;
   if(wdata->local == 0 && ftp_chdir(wdata, wdata->host, tempfle->file)) {
      update_ftp_info(wdata);
      gtk_clist_freeze(GTK_CLIST(wdata->listbox));
      delete_ftp_file_info(wdata);
      ftp_list_files(wdata, 1);
      gtk_clist_thaw(GTK_CLIST(wdata->listbox));
   }
   else if(wdata->local == 1) {
      if(!chdir(tempfle->file)) {
         ftp_log(LOG_MISC, "Successfully changed local directory to %s\n", tempfle->file);
         update_ftp_info(wdata);
         gtk_clist_freeze(GTK_CLIST(wdata->listbox));
         delete_ftp_file_info(wdata);
         add_local_files(wdata);
         gtk_clist_thaw(GTK_CLIST(wdata->listbox));
      }
      else {
         ftp_log(LOG_MISC, "Could not change %s directory to %s: %s\n", 
            wdata->local == 1 ? "local" : "remote", tempfle->file, g_strerror(errno));
      }
   }
}
/*****************************************************************************/
void chdiredit(GtkWidget *widget, struct ftp_window_data *wdata) {
   char tempstr[MAXSTR];
   
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Chdir: Not connected to a remote site\n");
      return;
   }
   if(!expand_path(gtk_entry_get_text(GTK_ENTRY(wdata->diredit)), tempstr, sizeof(tempstr))) return;
   if(wdata->local == 0 && ftp_chdir(wdata, wdata->host, tempstr)) {
      update_ftp_info(wdata);
      gtk_clist_freeze(GTK_CLIST(wdata->listbox));
      delete_ftp_file_info(wdata);
      ftp_list_files(wdata, 1);
      gtk_clist_thaw(GTK_CLIST(wdata->listbox));
   }
   else if(wdata->local == 1) {
      if(!chdir(tempstr)) {
         ftp_log(LOG_MISC, "Successfully changed local directory to %s\n", tempstr);
         update_ftp_info(wdata);
         gtk_clist_freeze(GTK_CLIST(wdata->listbox));
         delete_ftp_file_info(wdata);
         add_local_files(wdata);
         gtk_clist_thaw(GTK_CLIST(wdata->listbox));
      }
      else {
         ftp_log(LOG_MISC, "Could not change %s directory to %s: %s\n", 
            wdata->local == 1 ? "local" : "remote", tempstr, g_strerror(errno));
      }
   }
}
/*****************************************************************************/
void refresh(struct ftp_window_data *wdata) {
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Refresh: Not connected to a remote site\n");
      return;
   }
   gtk_clist_freeze(GTK_CLIST(wdata->listbox));
   delete_ftp_file_info(wdata);
   if(wdata->local == 0) ftp_list_files(wdata, 0);
   else add_local_files(wdata);
   gtk_clist_thaw(GTK_CLIST(wdata->listbox));
   update_ftp_info(wdata);
}
/*****************************************************************************/
void exitCB(gpointer data) {
   if(file_transfers == NULL) doexit(NULL, NULL);
   else {
      MakeYesNoDialog("Exit", "There are file transfers in progress.\nAre you sure you want to exit?", 2,
         "Exit", doexit, NULL,
         "Don't Exit", NULL, NULL);
   }
}
/*****************************************************************************/
void doexit(GtkWidget *widget, gpointer data) {
   clear_cache_files();
   gtk_main_quit();
}
/*****************************************************************************/
void disconnect(struct ftp_window_data *wdata) {
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Disconnect: Not connected to a remote site\n");
      return;
   }
   ftp_disconnect(wdata, wdata->host, 1);
}
/*****************************************************************************/
void savelog(gpointer data) {
   GtkWidget *filew;
   
   filew = gtk_file_selection_new ("Save Log");

   gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked", GTK_SIGNAL_FUNC(dosavelog), filew);
   gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button), "clicked", GTK_SIGNAL_FUNC(savelog_destroy), filew);
   gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked", GTK_SIGNAL_FUNC(savelog_destroy), filew);
   gtk_file_selection_set_filename(GTK_FILE_SELECTION(filew), "gftp.log");
   gtk_widget_show(filew);
}
/*****************************************************************************/
void dosavelog(GtkWidget *widget, GtkFileSelection *fs) {
   char *filename, *txt;
   guint textlen;
   FILE *fd;

   filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION (fs));
   fd = fopen(filename, "w");
   if(fd == NULL) {
      ftp_log(LOG_MISC, "Error: Cannot open %s: %s\n", filename, g_strerror(errno));
      return;
   }
   textlen = gtk_text_get_length(GTK_TEXT(logwdw));
   txt = gtk_editable_get_chars(GTK_EDITABLE(logwdw), 0, -1);
   if(!fwrite(txt, 1, textlen, fd)) {
      ftp_log(LOG_MISC, "Error: Error writing to %s\n", filename);
   }
   else {
      ftp_log(LOG_MISC, "Successfully wrote the log file to %s\n", filename);
   }
   fclose(fd);
   free(txt);
}
/*****************************************************************************/
void savelog_destroy(GtkWidget *widget, GtkWidget *dialog) {
   gtk_widget_destroy(dialog);
}
/*****************************************************************************/
void clearlog(gpointer data) {
   guint pos;
   
   pos = gtk_text_get_length(GTK_TEXT(logwdw));
   gtk_text_set_point(GTK_TEXT(logwdw), pos);
   gtk_text_backward_delete(GTK_TEXT(logwdw), pos);
}
/*****************************************************************************/
void fix_display(void) {
   while(gtk_events_pending()) gtk_main_iteration();
}
/*****************************************************************************/
gint update_downloads(gpointer data) {
   char *add_data[3] = {NULL, NULL, NULL}, *pos;
   struct ftp_transfer_data *temptdata, *prevtdata;
   struct ftp_log_queue *templog;
   char tempstr[MAXSTR];
   struct ftp_window_data *tempwdata;
   pthread_t tid;
   gint num, flag;
   
   if(file_transfer_logs != NULL) {
      while(file_transfer_logs != NULL) {
         templog = file_transfer_logs;
         file_transfer_logs = file_transfer_logs->next;
         ftp_log(templog->type, templog->msg);
         free(templog);
      }
   }
   if(file_transfers == NULL) {
      gtk_timeout_add(1000, update_downloads, data);
      return(0);
   }
   else {
      temptdata = prevtdata = file_transfers;
      num = 0;
      pthread_mutex_lock(&transfer_mutex);
      while(temptdata != NULL) {
         if(temptdata->flags & TRANSFER_SHOW) {
            g_snprintf(temptdata->progressstr, sizeof(temptdata->progressstr), "Waiting... (%d file%s)", 
               temptdata->hdata->totalfiles, temptdata->hdata->totalfiles > 1 ? "s" : "");
            temptdata->progressstr[sizeof(temptdata->progressstr)-1] = '\0';
            gtk_clist_insert(GTK_CLIST(dlwdw), num,
               add_data);
            temptdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
            temptdata->flags |= TRANSFER_UPDATE | TRANSFER_CREATED;
            temptdata->curfle = temptdata->hdata->files;
         }
         else if(temptdata->flags & TRANSFER_DONE) {
            tempwdata = (temptdata->flags & TRANSFER_DIRECTION) ? &window1 : &window2;
            if((temptdata->flags & TRANSFER_DIRECTION) || compare_hdata_structs(temptdata->hdata, tempwdata->host)) {
               refresh(tempwdata);
            }
            if(temptdata->flags & TRANSFER_CREATED) {
               /* We have to release the mutex because when the list item gets
                  deleted, it will call my deselect function. This func will
                  also lock the mutex */
               pthread_mutex_unlock(&transfer_mutex);
               gtk_clist_remove(GTK_CLIST(dlwdw), num);
               pthread_mutex_lock(&transfer_mutex);
            }
            free(temptdata->hdata);

            if(temptdata->flags & TRANSFER_DONE_DO_VIEW) view_file(temptdata->dest, 1);

            if(temptdata == prevtdata) {
               flag = 1;
               file_transfers = file_transfers->next;
            }
            else {
               flag = 0;
               prevtdata->next = temptdata->next;
            }
            free(temptdata);
            if(flag) temptdata = prevtdata = file_transfers;
            else temptdata = prevtdata->next;
            if(do_one_transfer_at_a_time || file_transfers == NULL) transfer_in_progress = 0;
            continue;
         }

         if(temptdata->flags & TRANSFER_UPDATE) {
            if(!(temptdata->flags & TRANSFER_STARTED) && start_file_transfers && (!do_one_transfer_at_a_time || (do_one_transfer_at_a_time && !transfer_in_progress))) {
               transfer_in_progress = 1;
               temptdata->flags |= TRANSFER_STARTED;
               if(temptdata->flags & TRANSFER_DIRECTION) pthread_create(&tid, NULL, &ftp_get_files, temptdata);  
               else pthread_create(&tid, NULL, &ftp_put_files, temptdata);
            }            

            pos = strrchr(temptdata->curfle->file, '/');
            if(pos == NULL) pos = temptdata->curfle->file;
            else pos++;
            gtk_clist_set_text(GTK_CLIST(dlwdw), num, 0, pos);

            gtk_clist_set_text(GTK_CLIST(dlwdw), num, 1, temptdata->progressstr);

            g_snprintf(tempstr, sizeof(tempstr), "%s%s", temptdata->hdata->host, temptdata->hdata->dir);
            tempstr[sizeof(tempstr)-1] = '\0';
            gtk_clist_set_text(GTK_CLIST(dlwdw), num, 2, tempstr);
         }
         num++;
         prevtdata = temptdata;
         temptdata = temptdata->next;
      }
      pthread_mutex_unlock(&transfer_mutex);
   }
   gtk_timeout_add(1000, update_downloads, data);
   return(0);
}
/*****************************************************************************/
void selectdl(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data) {
   struct ftp_transfer_data *tdata;
   int i;
   
   tdata = file_transfers;
   for(i=0; i<row; i++) tdata = tdata->next;
   pthread_mutex_lock(&transfer_mutex);
   tdata->flags |= TRANSFER_SELECTED;
   pthread_mutex_unlock(&transfer_mutex);
}
/*****************************************************************************/
void unselectdl(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data) {
   struct ftp_transfer_data *tdata;
   int i;
   
   tdata = file_transfers;
   for(i=0; i<row; i++) tdata = tdata->next;
   pthread_mutex_lock(&transfer_mutex);
   tdata->flags &= ~TRANSFER_SELECTED;
   pthread_mutex_unlock(&transfer_mutex);
}
/*****************************************************************************/
void stop_transfer(GtkWidget *widget, gpointer data) {
   struct ftp_transfer_data *temptdata;

   if(file_transfer_logs != NULL) {
      ftp_log(LOG_MISC, "There are currently no file transfers in progress to stop\n");
      return;
   }
   pthread_mutex_lock(&transfer_mutex);
   temptdata = file_transfers;
   while(temptdata != NULL) {
      if(temptdata->flags & TRANSFER_SELECTED) {
         temptdata->flags &= ~(TRANSFER_NEW | TRANSFER_SHOW | TRANSFER_UPDATE | TRANSFER_DONE);
         temptdata->flags |= TRANSFER_CANCEL;
         ftp_log(LOG_MISC, "Stopping the transfer of %s\n", temptdata->curfle->file);
      }
      temptdata = temptdata->next;
   }
   pthread_mutex_unlock(&transfer_mutex);
}
/*****************************************************************************/
int compare_hdata_structs(struct ftp_host_data *hdata1, struct ftp_host_data *hdata2) {
#ifdef DEBUG
   printf("Comparing: %ld %ld\n", hdata1->port, hdata2->port);
   printf("Comparing: %s %s\n", hdata1->host, hdata2->host);
   printf("Comparing: %s %s\n", hdata1->user, hdata2->user);
   printf("Comparing: %s %s\n", hdata1->passwd, hdata2->passwd);
   printf("Comparing: %s %s\n", hdata1->dir, hdata2->dir);
#endif
   if(hdata1->port == hdata2->port && strcmp(hdata1->host, hdata2->host) == 0 &&
      strcmp(hdata1->user, hdata2->user) == 0 && strcmp(hdata1->passwd, hdata2->passwd) == 0 &&
      strcmp(hdata1->dir, hdata2->dir) == 0) return(1);
   else return(0);
}
/*****************************************************************************/
void compare_windows(gpointer data) {
   struct ftp_file_data *curfle, *tempfle;
   gint num;

   if(window2.local == -1) {
      ftp_log(LOG_MISC, "Compare Windows: Not connected to a remote site\n");
      return;
   }
   
   deselectall(&window1);
   deselectall(&window2);
   
   /* Select the items in Window1 that aren't in Window2 */
   curfle = window1.host->files;
   num = 0;
   while(curfle != NULL) {
      tempfle = window2.host->files;
      if(!(curfle->flags & FILE_ISDIR)) {
         while(tempfle != NULL) {
            if(strcmp(tempfle->file, curfle->file) == 0 && tempfle->size == curfle->size) {
               break;
            }
            tempfle = tempfle->next;
         }
      }
      if(tempfle == NULL) gtk_clist_select_row(GTK_CLIST(window1.listbox), num, 0);
      num++;
      curfle = curfle->next;
   }

   /* Select the items in Window2 that aren't in Window1 */
   curfle = window2.host->files;
   num = 0;
   while(curfle != NULL) {
      tempfle = window1.host->files;
      if(!(curfle->flags & FILE_ISDIR)) {
         while(tempfle != NULL) {
            if(strcmp(tempfle->file, curfle->file) == 0 && tempfle->size == curfle->size) {
               break;
            }
            tempfle = tempfle->next;
         }
      }
      if(tempfle == NULL) gtk_clist_select_row(GTK_CLIST(window2.listbox), num, 0);
      num++;
      curfle = curfle->next;
   }
}
/*****************************************************************************/
